import numpy as np
import glob
import pandas as pd
import os
import time
from tqdm.auto import tqdm

from .misc import dfMirror

########################################################################################################################

# it's best to use asciiToDfMulti() (which exploits this asciiToDf()) also for single file opening
def asciiToDf(
        nameFormat,
        asciiMap,
        nLinesEv = 1,
        descFrac = 1,
        mirrorMap = (),  # this is a tuple here, but a dictionary in asciiToDfMulti() (i.e. the "main" function)
        bVerbose = False,
        bProgress = False,
):

    t0 = time.time()  # chronometer start
    names = sorted(glob.glob(nameFormat.replace("YYYYYY", "*")))  # list of all the filenames of the current run
    df = pd.DataFrame()
    descFrac = 1e-12 if descFrac <= 0 else (descFrac if descFrac <= 1 else 1)
    for iName in tqdm((names)) if (bVerbose & bProgress) else names:
        if os.stat(iName).st_size > 0:
            if nLinesEv == 1:
                dataTableTemp = np.loadtxt(iName, unpack=False, ndmin=2)
            else:
                fileToString0 = open(iName,'r').read()
                fileToStringSplitted0 = fileToString0.splitlines()
                fileToString = ""
                for i, iLine in enumerate(fileToStringSplitted0):
                    if (i%nLinesEv==nLinesEv-1):
                        fileToString += iLine + "\n"
                    else:
                        fileToString += iLine + " "
                fileToStringSplitted = fileToString.splitlines()
                dataTableTemp = np.loadtxt(fileToStringSplitted)
            dfTemp = pd.DataFrame(dataTableTemp, columns=asciiMap)
            df = df.append(dfTemp[dfTemp.index % int(1 / descFrac) == 0], ignore_index=True, sort=False)
            df = dfMirror(df, mirrorMap)
    t1 = time.time()  # chronometer stop
    dt = t1 - t0
    return df, dt

########################################################################################################################

def asciiToDfMulti(
        nameFormat,
        fileIndex,
        asciiMap,
        fileIndexName = "iIndex",
        nLinesEv = 1,
        descFrac = {},
        mirrorMap = {},
        bVerbose = False,
        bProgress = False,
):

    t0 = time.time()  # chronometer start
    df = pd.DataFrame()
    for i, iIndex in enumerate(sorted(fileIndex)):
        if not (iIndex in descFrac.keys()):
            descFrac.update({iIndex: 1})  # all the undefined descaling factors are trivially set to 1
        if bVerbose:
            print("(%d/%d) %s -- descaling fraction: %14.12f" % (i+1, len(fileIndex), iIndex, descFrac[iIndex]))
        dfTemp, _ = asciiToDf(nameFormat.replace("XXXXXX", iIndex), asciiMap, nLinesEv, descFrac[iIndex], bVerbose=bVerbose, bProgress=bProgress)

        # data mirroring according to mirrorMap, which differs from iLayer to iLayer
        if iIndex in mirrorMap:
            if bVerbose:
                print("mirroring (from mirror map given) "+str(mirrorMap[iIndex]))
            dfTemp = dfMirror(dfTemp, mirrorMap[iIndex])
        else:
            if bVerbose:
                print("no variables to mirror")

        # fileIndexName column creation (if requested & not already existing)
        if len(fileIndexName)>0:
            if not (fileIndexName in dfTemp.columns):
                dfTemp[fileIndexName] = str(iIndex)
            else:
                dfTemp[fileIndexName] = dfTemp[fileIndexName].astype(str)
            if bVerbose:
                print("%s also added to df" % fileIndexName)

        df = df.append(dfTemp, ignore_index=True, sort=False)
    t1 = time.time()  # chronometer stop
    dt = t1 - t0
    return df, dt



import numpy as np

########################################################################################################################

def fMCS(
        E,  # in GeV
        x,
        X0 = 1,
        z = 1,
        bLogTerm = True  # if True (False) logarithmic term is (not) included
):

    eRatio = 13.6e-3 / E
    linTerm = eRatio * z * np.sqrt(x/X0) * 1
    logTerm = eRatio * z * np.sqrt(x/X0) * 0.038*np.log(x/X0)
    return (linTerm + logTerm) if bLogTerm else linTerm


import numpy as np
import matplotlib.pyplot as plt

########################################################################################################################

def hist2dRatio(
        xNum, yNum,
        xDen, yDen,
        bins=None,
        range=None,
        bPlot=True,
        ax=None,
        norm=None,
        cmap=None,
):

    bins = bins if not (bins is None) else 10

    # denominator
    histDen = np.histogram2d(xDen, yDen, bins=bins, range=range)

    # if range is None, it is set from histDen
    if range is None:
        range = [[histDen[1][0], histDen[1][len(histDen[1])-1]], [histDen[2][0], histDen[2][len(histDen[2])-1]]]

    # numerator
    histNum = np.histogram2d(xNum, yNum, bins=bins, range=range)

    # ratio
    histDen[0][histDen[0] == 0] = np.nan  # replace all zeros in the denominator with NaNs
    histRatio = np.transpose(histNum[0] / histDen[0])

    if bPlot:  # output plot only if requested (default)
        histRatioPlot = np.nan_to_num(histRatio, nan=0.0)  # replace all the NaNs in the ratio with zeros, just for plot drawing
        if ax!=None:
            ax.imshow(histRatioPlot, origin="lower", extent=[min(histDen[1]), max(histDen[1]), min(histDen[2]), max(histDen[2])], norm=norm, cmap=cmap, aspect="auto")
        else:
            plt.imshow(histRatioPlot, origin="lower", extent=[min(histDen[1]), max(histDen[1]), min(histDen[2]), max(histDen[2])], norm=norm, cmap=cmap, aspect="auto")

    return histRatio, histDen[1], histDen[2]  # output matrix is always computed and returned



import pandas as pd
import uproot
import time
import glob
from tqdm.auto import tqdm

from .misc import dfReshape, dfMirror

########################################################################################################################

def rootToDfMulti(
        nameFormat,
        fileIndex,
        treeName,
        fileIndexName = "iIndex",
        descFrac = {},
        treeMap = {},
        mirrorMap = {},
        bVerbose = False,
        bProgress = False,
):

    t0 = time.time()  # chronometer start
    df = pd.DataFrame()
    for i, iIndex in enumerate(sorted(fileIndex)):
        names = sorted(glob.glob(nameFormat.replace("XXXXXX", iIndex).replace("YYYYYY", "*")))  # list of all the filenames of the current run
        if not (iIndex in descFrac.keys()):
            descFrac.update({iIndex: 1})  # all the undefined descaling factors are trivially set to 1
        descFrac[iIndex] = 1e-12 if descFrac[iIndex] <= 0 else (descFrac[iIndex] if descFrac[iIndex] <= 1 else 1)

        dfTemp = pd.DataFrame()
        if bVerbose:
            print("(%d/%d) %s -- descaling fraction: %14.12f" % (i + 1, len(fileIndex), iIndex, descFrac[iIndex]))
        for iName in tqdm((names)) if (bVerbose & bProgress) else names:  # for each value of iIndex, look for all the corresponding files
            tree = uproot.open(iName)[treeName]
            dfTemp0 = tree.arrays(library="pd")
            dfTemp = dfTemp.append(dfTemp0[dfTemp0.index % int(1 / descFrac[iIndex]) == 0], ignore_index=True, sort=False)

        # data reshaping: removing the square brackets in the names & remapping all the names according to treeMap
        if len(treeMap)>0:
            if bVerbose:
                print("remapping some ROOT tree variables (from tree map given)")
            if dfTemp.shape[0] > 0:
                dfTemp = dfReshape(dfTemp, treeMap, True)

        # data mirroring according to mirrorMap, which differs from iLayer to iLayer
        if iIndex in mirrorMap:
            if bVerbose:
                print("mirroring (from mirror map given) "+str(mirrorMap[iIndex]))
            if dfTemp.shape[0] > 0:
                dfTemp = dfMirror(dfTemp, mirrorMap[iIndex])
        else:
            if bVerbose:
                print("no variables to mirror")

        # fileIndexName column creation (if requested & not already existing -- after the data reshaping)
        if len(fileIndexName)>0:
            if bVerbose:
                print("%s also added to df" % fileIndexName)
            if dfTemp.shape[0] > 0:
                if not (fileIndexName in dfTemp.columns):
                    dfTemp[fileIndexName] = str(iIndex)
                else:
                    dfTemp[fileIndexName] = dfTemp[fileIndexName].astype(str)

        df = df.append(dfTemp, ignore_index=True, sort=False)
    t1 = time.time()  # chronometer stop
    dt = t1 - t0
    return df, dt



########################################################################################################################

def dfMirror(
        df,
        map
):

    for iVar in map:
        if iVar in df.columns:
            df[iVar] = -df[iVar]
    return df

########################################################################################################################

def dfReshape(
        df,
        map,
        bBrackets = True,
):

    # remove square brackets from the variable names (if required)
    if bBrackets:
        df = df.rename(columns = dict(zip(
            [s for s in df.columns if ("[" in s) & ("]" in s)],
            [s.replace("[", "").replace("]", "") for s in df.columns if ("[" in s) & ("]" in s)]
        )))

    # rename variables according to map
    df = df.rename(columns = dict(zip(
        [map[s] for s in map if map[s] in df.columns],
        [s for s in map if map[s] in df.columns]
    )))
    return df



import numpy as np

########################################################################################################################

def hist2dToProfile(
        hist2d,
        errType = ""
):

    xVal = []
    yVal = []
    yErr = []
    for i in range(len(hist2d[1])-1):
        yNum = 0
        yErrNum = 0
        yDenom = 0
        for j in range(len(hist2d[2])-1):
            yNum += hist2d[0][i][j] * (hist2d[2][j] + (hist2d[2][j+1] - hist2d[2][j])/2)
            yDenom += hist2d[0][i][j]
        if yDenom != 0:
            for j in range(len(hist2d[2])-1):
                yErrNum += hist2d[0][i][j] * \
                           ((hist2d[2][j] + (hist2d[2][j + 1] - hist2d[2][j]) / 2) - yNum/yDenom) ** 2
            yErrDenom = yDenom - 1
            xVal.append(hist2d[1][i] + (hist2d[1][i+1] - hist2d[1][i])/2)
            yVal.append(yNum/yDenom)
            if yDenom > 1:
                if errType == "std":
                    yErr.append(np.sqrt(yErrNum/yErrDenom))
                elif errType == "mean":
                    yErr.append(np.sqrt(yErrNum/yErrDenom) / np.sqrt(yDenom))
                else:
                    yErr.append(0)
            else:
                yErr.append(0)
    return np.array(xVal), np.array(yVal), np.array(yErr)



from setuptools import setup, find_packages

version = "2021.1.3"

setup(
    author="Mattia Soldani",
    author_email="mattiasoldani93@gmail.com",
    name="succolib",
    version=version,
    url="https://github.com/mattiasoldani/succolib",
    download_url = "https://github.com/mattiasoldani/succolib/archive/v"+version+".tar.gz",
    description="A set of handy, Python-based tools for the INSULAb detectors data analysis",
    long_description="This is **succolib**, a library of handy Python functions for High-Energy Physics beamtests data analysis. In particular, it has been developed with a focus on the event-by-event analysis of the data collected with the INSULAb detectors &mdash; see, for example, the experimental configurations described [here](http://cds.cern.ch/record/2672249), [here](http://hdl.handle.net/10277/857) and [here](http://cds.cern.ch/record/1353904). Details can be found in the [succolib GitHub page](https://github.com/mattiasoldani/succolib).",
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">3.0",
    install_requires=[
        "numpy",
        "pandas",
        "tqdm",
        "uproot>=4.0",
    ],
    license="MIT",
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Physics",
        "Intended Audience :: Science/Research",
    ],
)



import numpy as np
import time

########################################################################################################################

def eventSmear(
    dfIn,
    lsVar,
    lsSigma,
    nIter,
    bSmearSingleIter=False,
    bKeepOld=False,
    bVerbose=False
):
    
    t0 = time.time()  # chronometer start
    
    # check whether the input arguments make sense --> if not, returns empty dictionary
    if len(lsVar) != len(lsSigma):  # lsVar must have the same length as lsSigma
        if bVerbose:
            print ("list of variables and list of corresponding errors have different length --> operation not performed")
        t1 = time.time()  # chronometer stop
        dt = t1 - t0
        return {}, dt
    
    for s in lsVar+lsSigma:  # all the variables in lsVar+lsSigma must be available in the input dataframe
        if not (s in dfIn.columns):
            if bVerbose:
                print("variable %s not in input dataframe --> operation not performed")
            t1 = time.time()  # chronometer stop
            dt = t1 - t0
            return {}, dt
    
    # setting up input dataframe properly
    # events in which at least 1 of the variables to be studied are NaN are excluded
    dfIn = dfIn[lsVar+lsSigma].dropna()
    
    ind = list(dfIn.index)
    var = {}
    for iVar in lsVar+lsSigma:  # each dfIn variable into a list; lsVar-then-lsSigma; both elements of lsVar & lsSigma are included here
        var.update({iVar: list(dfIn[iVar])})
    
    # preparing the output dictionary (original index & values included only if bKeepOld is True)
    dictOut = {"old_index": np.ndarray(shape=nIter*dfIn.count()[0])} if bKeepOld else {}
    for iVar in lsVar:
        dictOut.update({iVar: np.ndarray(shape=nIter*dfIn.count()[0])})
        if bKeepOld:
            dictOut.update({"old_"+iVar: np.ndarray(shape=nIter*dfIn.count()[0])})

    for i in range(len(ind)):  # loop over original events
        
        # retrieving the event central value & sigma for each variable
        means, stds = [], []     
        for iVar in lsVar+lsSigma:
            if iVar in lsVar:
                means.append(var[iVar][i])
            elif iVar in lsSigma:
                stds.append(var[iVar][i])
    
        # gaussian doping
        covMatr = np.zeros((len(lsVar), len(lsVar)), float)   
        if (nIter > 1) | ((nIter == 1) & bSmearSingleIter):
        # note: if nIter < 1 or =1 in case bSmearSingleIter=True, null covariance matrix is used --> output data equal the input ones
            np.fill_diagonal(covMatr, np.array(stds)**2)
        outStat = np.random.multivariate_normal(means, covMatr, nIter).T
        
        # filling the output dictionary
        for k, iVar in enumerate(lsVar):
            dictOut[iVar][i*nIter:(i+1)*nIter] = outStat[k]
            if bKeepOld:
                dictOut["old_"+iVar][i*nIter:(i+1)*nIter] = np.array([means[k] for j in range(nIter)])
                dictOut["old_index"][i*nIter:(i+1)*nIter] = np.array([ind[i] for j in range(nIter)])

    t1 = time.time()  # chronometer stop
    dt = t1 - t0
    return dictOut, dt